from __future__ import annotations
import numpy as np
from dataclasses import dataclass

TYPE_CHECKING = False
if TYPE_CHECKING:
    from typing import Any, TypeAlias, Sequence
    from numpy.typing import NDArray

    _DictKey: TypeAlias = str | int | tuple["_DictKey", ...]


def normalize_result_axis(
    axis: tuple[int, ...] | int | None,
    result: NDArray[Any],
    ell: Sequence[Any] | NDArray[Any] | None,
) -> tuple[int, ...]:
    """
    Normalize the axis argument for result arrays.

    Returns a tuple of axis indices, handling None and negative values.
    """
    try:
        from numpy.lib.array_utils import normalize_axis_tuple
    except ModuleNotFoundError:
        from numpy.lib.stride_tricks import normalize_axis_tuple  # type: ignore

    if axis is None:
        if result.ndim == 0:
            axis = ()
        elif isinstance(ell, tuple):
            axis = tuple(range(-len(ell), 0))
        else:
            axis = -1
    return normalize_axis_tuple(axis, result.ndim, "axis")


@dataclass(frozen=True, repr=False)
class AngularPowerSpectrum:
    """
    AngularPowerSpectrum: A sleek dataclass for LSS numerical results and metadata.
    for angular power spectra and mixing matrix results.

    Attributes:
        array: Main power spectrum data (float dtype).
        ell: Multipole moment(s) or bin centers.
        axis: Axis or axes corresponding to result dimensions.
        lower: Optional lower error bounds.
        upper: Optional upper error bounds.
        weight: Optional weights (e.g. mode counts or covariance weights).
        software: Optional software identifier.
    """

    array: NDArray[Any]
    ell: NDArray[Any] | tuple[NDArray[Any], ...] | None = None
    axis: int | tuple[int, ...] | None = None
    lower: NDArray[Any] | tuple[NDArray[Any], ...] | None = None
    upper: NDArray[Any] | tuple[NDArray[Any], ...] | None = None
    weight: NDArray[Any] | tuple[NDArray[Any], ...] | None = None
    software: str | None = None
    metadata: dict[str, Any] | None = None

    def __post_init__(self) -> None:
        float_array = np.asarray(self.array, dtype=float)
        object.__setattr__(self, "array", float_array)
        axis = normalize_result_axis(self.axis, self.array, self.ell)
        object.__setattr__(self, "axis", axis)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(axis={self.axis!r})"

    def __array__(
        self,
        dtype: np.dtype[Any] | None = None,
        *,
        copy: bool | None = None,
    ) -> NDArray[Any]:
        if copy is not None:
            return self.array.__array__(dtype, copy=copy)
        return self.array.__array__(dtype)

    def __getitem__(self, key: Any) -> Any:
        return self.array[key]

    @property
    def ndim(self) -> int:
        return self.array.ndim

    @property
    def shape(self) -> tuple[int, ...]:
        return self.array.shape

    @property
    def dtype(self) -> np.dtype[Any]:
        return self.array.dtype


@dataclass(frozen=True, repr=False)
class TwoPointCorrelationFunction:
    """
    TwoPointCorrelationFunction: Dataclass for 2PCF results and metadata.

    Attributes:
        array: Main correlation data (float dtype).
        theta: Angular separation(s) or bin centers.
        axis: Axis or axes corresponding to result dimensions.
        lower: Optional lower error bounds.
        upper: Optional upper error bounds.
        weight: Optional weights (e.g. pair counts or covariance weights).
        software: Optional software identifier.
    """

    array: NDArray[Any]
    theta: NDArray[Any] | tuple[NDArray[Any], ...] | None = None
    axis: int | tuple[int, ...] | None = None
    lower: NDArray[Any] | tuple[NDArray[Any], ...] | None = None
    upper: NDArray[Any] | tuple[NDArray[Any], ...] | None = None
    weight: NDArray[Any] | tuple[NDArray[Any], ...] | None = None
    software: str | None = None

    def __post_init__(self) -> None:
        float_array = np.asarray(self.array, dtype=float)
        object.__setattr__(self, "array", float_array)
        axis = normalize_result_axis(self.axis, self.array, self.theta)
        object.__setattr__(self, "axis", axis)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(axis={self.axis!r})"

    def __array__(
        self,
        dtype: np.dtype[Any] | None = None,
        *,
        copy: bool | None = None,
    ) -> NDArray[Any]:
        if copy is not None:
            return self.array.__array__(dtype, copy=copy)
        return self.array.__array__(dtype)

    def __getitem__(self, key: Any) -> Any:
        return self.array[key]

    @property
    def ndim(self) -> int:
        return self.array.ndim

    @property
    def shape(self) -> tuple[int, ...]:
        return self.array.shape

    @property
    def dtype(self) -> np.dtype[Any]:
        return self.array.dtype


@dataclass(frozen=True, repr=False)
class COSEBI:
    """
    AngularPowerSpectrum: A sleek dataclass for LSS numerical results and metadata.
    for angular power spectra and mixing matrix results.

    Attributes:
        array: Main power spectrum data (float dtype).
        axis: Axis or axes corresponding to result dimensions.
        mode: Multipole moment(s) or bin centers.
        nmode: Optional number of modes per bin
        thmin: Optional minimum angular scale.
        thmax: Optional maximum angular scale.
        software: Optional software identifier.
    """

    array: NDArray[Any]
    axis: int | tuple[int, ...] | None = None
    mode: int | tuple[int, ...] | None = None
    nmodes: int | tuple[int, ...] | None = None
    thmin: float | None = None
    thmax: float | None = None
    software: str | None = None

    def __post_init__(self) -> None:
        float_array = np.asarray(self.array, dtype=float)
        object.__setattr__(self, "array", float_array)
        axis = normalize_result_axis(self.axis, self.array, self.mode)
        object.__setattr__(self, "axis", axis)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(axis={self.axis!r})"

    def __array__(
        self,
        dtype: np.dtype[Any] | None = None,
        *,
        copy: bool | None = None,
    ) -> NDArray[Any]:
        if copy is not None:
            return self.array.__array__(dtype, copy=copy)
        return self.array.__array__(dtype)

    def __getitem__(self, key: Any) -> Any:
        return self.array[key]

    @property
    def ndim(self) -> int:
        return self.array.ndim

    @property
    def shape(self) -> tuple[int, ...]:
        return self.array.shape

    @property
    def dtype(self) -> np.dtype[Any]:
        return self.array.dtype



from __future__ import annotations
from dataclasses import dataclass

TYPE_CHECKING = False
if TYPE_CHECKING:
    from typing import Any
    from numpy.typing import NDArray


@dataclass(frozen=True)
class PowerSpectrumMultipoles:
    """
    Output data model to store power spectrum Legendre multipole measurements.

    Attributes
    ----------
    k : ndarray
        Centers of $k$ bins
    keff : ndarray
        Effective values of $k$ bins
    Nmodes : ndarray
        Number of modes in each $k$ bin
    multipoles : ndarray
        Legendre multipoles of the anisotropic power spectrum
    fiducial_cosmology : dict[str, float]
        Fiducial cosmology used in the measurement
    zeff : ndarray
        Effective redshift of the sample
    nbar : float
        Number density of the sample
    Psn : float
        Poissonian shot noise contribution
    """
    k: NDArray[Any]
    keff: NDArray[Any]
    Nmodes: NDArray[Any]
    multipoles: NDArray[Any]
    fiducial_cosmology: dict[str, float]
    zeff: float
    nbar: float
    Psn: float

    def __post_init__(self) -> None:
        if any([len(self.k) != len(self.keff),
                len(self.k) != len(self.Nmodes),]):
            raise ValueError(
                "Inconsistent class attributes, all arrays must have the same "
                " length.")

    def __len__(self) -> int:
        return len(self.k)


@dataclass(frozen=True)
class TwoPointCorrelationCartesian:
    """
    Output data model to store 2-point correlation function (cartesian) measurements.

    Attributes
    ----------
    s_perp : ndarray
        Separation $s_\perp$ ortoghonal to the line of sight
    s_para : ndarray
        Separation $s_\parallel$ parallel to the line of sight
    correlation : ndarray
        2D cartesian 2PCF
    fiducial_cosmology : dict[str, float]
        Fiducial cosmology used in the measurement
    zeff : ndarray
        Effective redshift of the sample
    """
    s_perp: NDArray[Any]
    s_para: NDArray[Any]
    correlation: NDArray[Any]
    fiducial_cosmology: dict[str, float]
    zeff: float

    def __len__(self) -> int:
        return (len(self.s_perp), len(self.s_para))


@dataclass(frozen=True)
class TwoPointCorrelationPolar:
    """
    Output data model to store 2-point correlation function (polar) measurements.

    Attributes
    ----------
    s : ndarray
        Separation $s$
    mu : ndarray
        Cosinus $\mu$ of the angle to the line of sight
    correlation : ndarray
        2D polar 2PCF
    fiducial_cosmology : dict[str, float]
        Fiducial cosmology used in the measurement
    zeff : ndarray
        Effective redshift of the sample
    """
    s: NDArray[Any]
    mu: NDArray[Any]
    correlation: NDArray[Any]
    fiducial_cosmology: dict[str, float]
    zeff: float

    def __len__(self) -> int:
        return (len(self.s), len(self.mu))


@dataclass(frozen=True)
class TwoPointCorrelationMultipoles:
    """
    Output data model to store 2-point correlation function Legendre multipole measurements.

    Attributes
    ----------
    s : ndarray
        Separation $s$
    multipoles : ndarray
        Legendre multipoles of the anisotropic 2PCF
    fiducial_cosmology : dict[str, float]
        Fiducial cosmology used in the measurement
    zeff : ndarray
        Effective redshift of the sample
    """
    s: NDArray[Any]
    multipoles: NDArray[Any]
    fiducial_cosmology: dict[str, float]
    zeff: float

    def __len__(self) -> int:
        return len(self.s)


@dataclass(frozen=True)
class PowerSpectrumMultipolesCovariance:
    """
    Output data model to store covariance matrices of the power spectrum
    Legendre multipole measurements.

    Attributes
    ----------
    k : ndarray
        Centers of $k$ bins
    cov : dict
        Auto and cross covariances of the Legendre multipoles of the
        power spectrum
    zeff : ndarray
        Effective redshift of the sample
    """
    k: NDArray[Any]
    covariance: dict[str, NDArray[Any]]
    zeff: float

    def __len__(self) -> int:
        return len(self.k)


@dataclass(frozen=True)
class TwoPointCorrelationMultipolesCovariance:
    """
    Output data model to store covariance matrices of the 2-point correlation
    Legendre
    multipole measurements.

    Attributes
    ----------
    s : ndarray
        Separation $s$
    cov : dict
        Auto and cross covariances of the Legendre multipoles of the 2PCF
    zeff : ndarray
        Effective redshift of the sample
    """
    s: NDArray[Any]
    covariance: dict[str, NDArray[Any]]
    zeff: float

    def __len__(self) -> int:
        return len(self.s)


@dataclass(frozen=True)
class PowerSpectrumMultipolesMixingMatrix:
    """
    Output data model to store mixing matrices of the power spectrum Legendre
    multipole measurements.

    Attributes
    ----------
    kout : ndarray
        Output wavemodes $k$
    kin : dict[ndarray]
        Input wavemodes $k$ per multipole
    mixing : dict[ndarray]
        Individual blocks of the mixing matrix
    """
    kout: NDArray[Any]
    kin: dict[Any, NDArray[Any]]
    mixing: dict[str, NDArray[Any]]
    zeff: float

    def __len__(self) -> int:
        return (len(self.kout), len(self.kin))



# photo/harmonic_space/__init__.py
from .photo import AngularPowerSpectrum, TwoPointCorrelationFunction, COSEBI
from .spectro import PowerSpectrumMultipoles, TwoPointCorrelationCartesian, TwoPointCorrelationMultipoles, TwoPointCorrelationPolar
from .spectro import PowerSpectrumMultipolesCovariance, TwoPointCorrelationMultipolesCovariance
from .spectro import PowerSpectrumMultipolesMixingMatrix

__all__ = [
    "AngularPowerSpectrum",
    "TwoPointCorrelationFunction",
    "COSEBI",
    "PowerSpectrumMultipoles",
    "PowerSpectrumMultipolesCovariance",
    "PowerSpectrumMultipolesMixingMatrix",
    "TwoPointCorrelationCartesian",
    "TwoPointCorrelationPolar",
    "TwoPointCorrelationMultipoles",
    "TwoPointCorrelationMultipolesCovariance",
]
